/**
 *
 * Copyright (C) 2013, 2014 Olivier Berthier
 *                    http://www.xunil.fr
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 **/

#include "delaunay.h"


listSimplex* delaunay(vertex* L, unsigned long faces, unsigned long iterations, double seuil,int optimise){
  if(L==NULL)
    return NULL;

  listSimplex* queue=newListSimplex(2*(sizeList(L)-4)+2);

  if(splitQuad(L,queue)){    
#if HEAP>0
    while(1){
      simplex* s = removeMin(queue);
#else
    while(queue->s!=NULL){
      simplex* s = queue->s;
#endif
      
      //on sort quand il n'y a plus rien à trianguler
      if(s==NULL || s->vertices == NULL)  
	break;  

      if(seuil || optimise)
      	if(s->hauteur <= seuil)
      	  break;


      if(faces && faces <= queue->size)
	break;

      
      splitSimplex(queue,s);


      
      if(iterations){
	iterations--;
	//on sort qand le nombre d'itérations demandé est atteint
	if(!iterations)
	  break;
      }
    }

  }

  return queue;
}

void delaunay_it(listSimplex* queue){
#if HEAP>0
  simplex* s = removeMin(queue);
    splitSimplex(queue,s);
#else
  if(queue->s!=NULL){
    simplex* s = queue->s;
    splitSimplex(queue,s);
  }
#endif
}

/**
 * \brief fonction qui va séparer le quadrilatère formé par les 4 premiers points de L
 * Les deux triangles sont ajoutés à q, chacun des triangles à la liste des points qu'ils contient (projeté vertical)
 */
int splitQuad(vertex* L, listSimplex* l){

  //initialisation : diviser le quad initial
  simplex* triangle1= newSimplex();//ABD
  simplex* triangle2= newSimplex();//BCD
  
  vertex* s=L;
  
  triangle1->sommets[0] = s;
  addRef(s,1);
  s=s->link->next;
  triangle1->sommets[1] = s;
  triangle2->sommets[0] = s;
  addRef(s,2);
  s=s->link->next;
  triangle2->sommets[1] = s;
  addRef(s,1);
  s=s->link->next;
  triangle1->sommets[2] = s;
  triangle2->sommets[2] = s;
  addRef(s,2);

  triangle1->voisins[0]=triangle2;

  triangle2->voisins[1]=triangle1;

  //insertion des points + calcul de la hauteur max
  vertex* v= NULL;
  s = s->link->next;

  while(s!=NULL){
    v = s->link->next;
    if(vertexInSimplex(triangle1,s)!=DEHORS)
      addPointInSimplex(triangle1,s);
    else    
      addPointInSimplex(triangle2,s);
    s=v;
  }
#if HEAP>0
  insert(l,triangle1);
  insert(l,triangle2);
#else
  //insertion dans la file de priorité
  insertSimplexInList(triangle1,l);
  insertSimplexInList(triangle2,l);
#endif
  return 1;
}


int checkDelaunayFull(listSimplex* l, int m, int v){
  //chequer la triangulation
  //prendre chaque simplex
  //regarder si pour tout ses voisins, le point p est à l'exterieur du cercle
  simplex* s = l->s;
  double d,r,x,y;
  int error = 0,i;
#if HEAP>0
  int n;
    for(n = 0; n<l->size;n++){
      simplex* ss=l->ss[n];
#else
  while(s!= NULL){
    simplex* ss = l->s;
#endif
    circumcircle(s,&r,&x,&y);
    while(ss!= NULL){
      i=0;
      if(m==2){
	d=sqrt( (ss->sommets[i]->coords[0]-x)*(ss->sommets[i]->coords[0]-x) + (ss->sommets[i]->coords[1]-y)*(ss->sommets[i]->coords[1]-y));
      
	if(!(fabs(d-r)<EPSILON) && d<r){
	  error++;
	  if(v){
	    printVertex(ss->sommets[i]);
	    printf("Is in :\n");
	    printSimplex(s);
	    printf("%lf < %lf (%lf)\n",d,r,d-r);
	  }
	}

	i=1;
	d=sqrt( (ss->sommets[i]->coords[0]-x)*(ss->sommets[i]->coords[0]-x) + (ss->sommets[i]->coords[1]-y)*(ss->sommets[i]->coords[1]-y));
      
	if(!(fabs(d-r)<EPSILON) && d<r){
	  error++;
	  if(v){
	    printVertex(ss->sommets[i]);
	    printf("Is in :\n");
	    printSimplex(s);
	    printf("%lf < %lf (%lf)\n",d,r,d-r);
	  }
	}

	i=2;
	d=sqrt( (ss->sommets[i]->coords[0]-x)*(ss->sommets[i]->coords[0]-x) + (ss->sommets[i]->coords[1]-y)*(ss->sommets[i]->coords[1]-y));
      
	if(!(fabs(d-r)<EPSILON) && d<r){
	  error++;
	  if(v){
	    printVertex(ss->sommets[i]);
	    printf("Is in :\n");
	    printSimplex(s);
	    printf("%lf < %lf (%lf)\n",d,r,d-r);
	  }
	}
      }else{
      


      
	if(InCircle(s->sommets[0],s->sommets[1],s->sommets[2],ss->sommets[0]) == DEDANS){
	  error ++;//TODO=>print s is in abc
	  if(v){
	    printVertex(ss->sommets[0]);
	    printf("Is in :\n");
	    printSimplex(s);
	  }
	}
      
	if(InCircle(s->sommets[0],s->sommets[1],s->sommets[2],ss->sommets[1]) == DEDANS){
	  error ++;//TODO=>print s is in abc
	  if(v){
	    printVertex(ss->sommets[0]);
	    printf("Is in :\n");
	    printSimplex(s);
	  }
	}

      
	if(InCircle(s->sommets[0],s->sommets[1],s->sommets[2],ss->sommets[2]) == DEDANS){
	  error ++;//TODO=>print s is in abc
	  if(v){
	    printVertex(ss->sommets[0]);
	    printf("Is in :\n");
	    printSimplex(s);
	  }
	}
      }
      ss=ss->prev;
    }
    s=s->prev;
  }
  if(error) return 0;
  return 1;
}
