/**
 *
 * Copyright (C) 2013, 2014 Olivier Berthier
 *                    http://www.xunil.fr
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 **/

/**
 * \file simplex.h
 * \brief GA Triangulation Delaunay
 * \author Luis Felipe Restrepo Botero, Olivier Berthier
 * \date décembre 2013
 *
 */
#ifndef SIMPLEX_H
#define SIMPLEX_H

#include "commun.h"
#include "stack.h"
#include "vertex.h"

/** 
 * \struct simplex
 * \brief Structure pour stocker un simplex. 
 */
typedef struct _simplex{
  /*! Sommets du simplex */
  struct _vertex* sommets[3];
  /*! Simplexs voisins */
  struct _simplex* voisins[3];
  /*! Liste des vertex situés à l'interieur du simplex */
  struct _vertex* vertices;
  /*! simplex précédent */
  struct _simplex* prev;
  /*! simplex suivant */
  struct _simplex* next;
  /*! Liste à laquelle le simplex appartient */
  struct _list* list;
  /*! sens de rotation des vertex du simplex */
  int sens;
  /*! plus grande hauteur du simplex */
  double hauteur;
  /*! Indice dans le tas */
  int index;
  /*! Pointeur vers la pile qui contient le simplex */
  struct _stack* stack;
  struct simplex *(*newSimplex)();
  
} simplex;

/** 
 * \struct listSimplex
 * \brief Structure pour stocker une liste de simplex. 
 */
typedef struct _list{
  /*! simplex de début de liste */
  simplex* s;
  /*! simplex de fin de liste */
  simplex* end;
  /*! Taille de la liste */
  unsigned int size;//pour le critère d'arrêt
#if HEAP>0
  /*! simplex de début de liste */
  simplex** ss;
  int capacity;
#endif
  simplex** T;
  struct listSimplex *(*newListSimplex)(int capacity);
} listSimplex;


/**
 * \fn simplex *newSimplex()
 * \return simplex initialisé
 * \brief Fonction d'initalisation d'une structure simplex.
 */
simplex *newSimplex();

/**
 * \fn void printSimplex(simplex* s)
 * \param s simplex à afficher
 * \brief procédure permettant d'afficher un simplex.
 */
void printSimplex(simplex* s);

/**
 * \fn int getIndice(simplex* s, struct _vertex* v)
 * \param s simplex dans lequelle on recherche l'indice de v.
 * \param v vertex dont on recher l'indice.
 * \return indice du vertex dans le simplex.
 * \brief Fonction qui permet de retrouver à quelle indice se trouve un vertex parmis les sommets d'un simplex.
 */
int getIndice(simplex* s, struct _vertex* v);

/**
 * \fn void findAndReplaceVoisin(simplex* a, simplex* b, simplex* c)
 * \param a simplex dans lequelle on effectue le remplacement.
 * \param b simplex recherché
 * \param c nouveau simplex qui prendra la place de b
 * \brief Procédure qui permet de rechercher et remplacer un simplex voisin dans un simplex.
 */
void findAndReplaceVoisin(simplex* a, simplex* b, simplex* c);

/**
 * \fn int splitSimplex(listSimplex* queue,simplex* s)
 * \param queue File de priorité
 * \param s Simplex à diviser
 * \return 1 si ok, 0 sinon
 * \brief Procédure qui va diviser un simplex en 3 par rapport à son vertex interne ayant la plus grande distance orthogonale par rapport au plan.
 */
int splitSimplex(listSimplex* queue,simplex* s);

/**
 * \fn int checkDelaunay(listSimplex* queue, simplex* s1,simplex* s2,struct _vertex* v,struct _stack** stack)
 * \param queue File de priorité
 * \param s1 Simplex que l'on va contrôler.
 * \param s2 Simplex adjacent contenant le vertex de contrôle (celui qui ne doit pas se trouver dans le cercle circonscrit)
 * \param v Vertex de contrôle.
 * \param stack Pile.
 * \return 0 s'il y a eu inversion, sinon 1
 * \brief Fonction qui effectue un contrôle de Delaunay sur un simplex par rapport à un simplex adjacent.
 */
int checkDelaunay(listSimplex* queue, simplex* s1,simplex* s2,struct _vertex* v,struct _stack** stack);

/**
 * \fn void addPointInSimplex(simplex* s, struct _vertex* v)
 * \param s Simplex auquel on ajoute le vertex v.
 * \param v Vertex à ajouter.
 * \brief Procédure permettant d'ajouter un vertex à la liste des vertex internes au simplex.
 */
void addPointInSimplex(simplex* s, struct _vertex* v);

/**
 * \fn int vertexInSimplex(simplex* s, struct _vertex* p)
 * \param s Simplex dans lequel on recherche le vertex
 * \param p Vertex recherché
 * \return DEDANS (1) si le vertex se trouve à l'interieur, SUR (2) si le vertex est sur le bord, DEHORS (0) sinon.
 */
int vertexInSimplex(simplex* s, struct _vertex* p);

/**
 * \fn double hauteur(simplex* s, struct _vertex* v)
 * \param s Simplex servant de plan
 * \param v Vertex dont on cherche à calculer la distance au plan
 * \return hauteur
 * \brief Fonction permettant de calculer la hauteur orthogonal d'un vertex par rapport à un plan.
 */
double hauteur(simplex* s, struct _vertex* v);

/**
 * \fn void circumcircle(simplex* s, double* radius, double* x, double* y )
 * \param s Simplex pour lequel on recherche le cercle circonscrit.
 * \param radius (ref) Rayon du cercle.
 * \param x abscisse du centre du cercle.
 * \param y ordonné du centre du cercle.
 * \brief Procédure permettant de rechercher les paramètres du cercle circonscrit à un simplex.
 */
void circumcircle(simplex* s, double* radius, double* x, double* y );

/**
 * \fn int left(simplex* s)
 * \param s Simplex testé
 * \return indice du point le plus à gauche
 * \brief Fonction qui permet de rechercher l'indice du point le plus à gauche parmis les sommets d'un vertex.
 */
int left(simplex* s);

/**
 * \fn int checkSimplexDirection(simplex* s)
 * \param s Simplex à tester.
 * \return CCW (1) si sens horaire inverse, sinon CW (2)
 * \brief Fonction permettant de tester le sens de rotation des vertex d'un simplex.
 */
int checkSimplexDirection(simplex* s);

/**
 * \fn listSimplex *newListSimplex()
 * \return listSimplex initialisé
 * \brief Fonction permettant d'initialiser une liste de simplex.
 */
listSimplex *newListSimplex(int capacity);


/**
 * \fn void printListSimplex(listSimplex* l);
 * \param l Liste de simplex à afficher.
 * \brief Procédure permettant d'afficher une liste de simplex.
 */
void printListSimplex(listSimplex* l);

/**
 * \fn void insertSimplexInList(simplex* s, listSimplex* l)
 * \param s Simplex à inserer.
 * \param l Liste de simplex.
 * \brief Procédure permettant d'ajouter un simplex à une liste de simplex. Le simplex ayant la hauteur la plus élevée sera toujours en tête de liste. Le simplex est inséré avant celui ayant une hauteur moins élevée.
 */
void insertSimplexInList(simplex* s, listSimplex* l);

/**
 * \fn void removeSimplexFromList(simplex* s, listSimplex* l)
 * \param s Simplex que l'on souhaite retirer.
 * \param l Liste dans laquel on recherche le simplex.
 * \brief Procédure permettant de retirer un simplex d'une liste.
 */
void removeSimplexFromList(simplex* s, listSimplex* l);

/**
 * \fn void removeHead(listSimplex* l)
 * \param l Liste de simplex
 * \brief Procédure permettant de retirer la tête d'une liste de simplex.
 */
void removeHead(listSimplex* l);

/**
 * \fn void deleteListSimplex(listSimplex* l)
 * \brief Procédure permettant de libérer la mémoire allouée pour une liste de simplex.
 * \param l Liste de simplex à supprimer.
 */
void deleteListSimplex(listSimplex* l);

/**
 * \fn void moveInList(simplex* s,listSimplex* l)
 * \brief Procédure permettant de déplacer un objet déjà présent dans la file de priorité.
 * \param s Simplex à déplacer
 * \param l File de priorité
 */
void moveInList(simplex* s,listSimplex* l);

/**
 * \fn void pushFront(simplex* s, listSimplex* l)
 * \brief Procédure permettant de déplacer un simplex depuis sa position dans la file de de priorité vers le début.
 * \param s Simplex à déplacer
 * \param l File de priorité
 */
void pushFront(simplex* s, listSimplex* l);

/**
 * \fn void pushBack(simplex* s, listSimplex* l)
 * \brief Procédure permettant de déplacer un simplex depuis sa position dans la file de de priorité vers la fin.
 * \param s Simplex à déplacer
 * \param l File de priorité
 */
void pushBack(simplex* s, listSimplex* l);

/**
 * \fn void insertBack(simplex* s, listSimplex* l)
 * \brief Procédure permettant de déplacer un simplex d'une liste en partant du début
 * \param s Simplex à déplacer
 * \param l File de priorité
 */
void insertBack(simplex* s, listSimplex* l);

/**
 * \fn void insertAfter(simplex* in, simplex* new, listSimplex* l)
 * \brief Procédure permettant d'ajouter un simplex juste derière un simplex présent dans une file de priorité
 * \param in Simplex déjà présent
 * \param new nouveau simplex
 * \param l liste
 */
void insertAfter(simplex* in, simplex* new, listSimplex* l);

#if HEAP>0
/**
 * \fn void heapify( listSimplex* l)
 */
void heapify( listSimplex* l) ;

/**
 *\fn void percolateDown( listSimplex* l, int index)
 */
void percolateDown( listSimplex* l, int index) ;

/**
 * \fn void percolateUp( listSimplex* l, int index)
 */
void percolateUp( listSimplex* l, int index);

/**
 * \fn void insert(listSimplex* l, simplex* s)
 */
void insert(listSimplex* l, simplex* s);

/**
 * \fn simplex* removeMin(listSimplex* l) 
 */
simplex* removeMin(listSimplex* l) ;

/**
 * \fn int minimum(double a, int indexa, double b, int indexb) 
 */
int minimum(double a, int indexa, double b, int indexb) ;

/**
 * \fn void swap(listSimplex *h, int index1, int index2)
 */
void swap(listSimplex *h, int index1, int index2) ;

/**
 * \fn void printHeap(listSimplex *h)
 */
void printHeap(listSimplex *h);

#endif
#endif //DELAUNAY_H
